#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : BaseModel
# @Time         : 2019-06-22 22:21
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

from pathlib import Path
from abc import abstractmethod
from tensorflow.python.keras.utils import plot_model as plot_model_keras
from tensorflow.python.keras.layers import Embedding


# import tensorflow as tf
# tf.keras.backend.clear_session()


class BaseModel(object):
    """
    https://github.com/ShawnyXiao/TextClassification-Keras
    """

    @abstractmethod
    def __init__(self, max_tokens=20000, maxlen=128, embedding_size=None, num_class=1, weights=None,
                 best_model_weights=None, **kwargs):
        self.max_tokens = max_tokens
        self.maxlen = maxlen
        # 6 * int(pow(feat.dimension, 0.25)
        self.embedding_size = embedding_size if embedding_size else min(50, (max_tokens + 1) // 2)  # 经验值

        self.num_class = num_class
        self.last_activation = 'softmax' if num_class > 1 else 'sigmoid'

        self.weights = weights
        self.best_model_weights = best_model_weights

        # self.model = None

    def __call__(self, summary=False, plot_model=False, dir='.', **kwargs):
        model = self.get_model()
        if self.best_model_weights is not None:
            model.load_weights(self.best_model_weights, by_name=True)
        if summary:
            model.summary()
        if plot_model:
            image_file = Path(dir) / ('%s.png' % self._class_name)
            image_file = image_file.absolute().__str__()
            try:  # 必须return才能显示图
                plot_model_keras(model, to_file=image_file, show_shapes=True, dpi=128)
            except Exception as e:
                print(e)
                print("brew install graphviz or apt-get install graphviz")
        return model

    @abstractmethod
    def get_model(self):
        # # TODO: 基模型直接编译
        # """build
        # self.model = ...
        # self.model.compile()
        # """
        pass

    def plot_show(self, image_file=None):
        if image_file is None:
            image_file = '%s.png' % self._class_name
        try:
            from IPython import display
            return display.Image(filename=image_file)
        except ImportError:
            pass

    @property
    def _class_name(self):
        return str(self).split(maxsplit=1)[0][10:]

    @property
    def embedding_layer(self):
        if self.weights is not None:
            return Embedding(*self.weights.shape, input_length=self.maxlen, weights=[self.weights], trainable=False)
        else:
            return Embedding(self.max_tokens, self.embedding_size, input_length=self.maxlen)

    # def __call__(self, summary=False, plot_model=False, dir='.', **kwargs):
    #     model = self.get_model()
    #     if self.best_model_weights is not None:
    #         model.load_weights(self.best_model_weights, by_name=True)
    #     if summary:
    #         model.summary()
    #     if plot_model:
    #         image_file = Path(dir) / ('%s.png' % self._class_name)
    #         image_file = image_file.absolute().__str__()
    #         try:  # 必须return才能显示图
    #             plot_model_keras(model, to_file=image_file, show_shapes=True, dpi=128)
    #         except Exception as e:
    #             print(e)
    #             print("brew install graphviz or apt-get install graphviz")
    #     return model

if __name__ == '__main__':
    bm = BaseModel()
    print(bm._class_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'bayes_xgb'
__author__ = 'JieYuan'
__mtime__ = '19-1-3'
"""
import warnings

warnings.filterwarnings('ignore')

import pandas as pd
import xgboost as xgb
from bayes_opt import BayesianOptimization
from bayes_opt.observer import JSONLogger
from bayes_opt.event import Events


class BayesOptXGB(object):
    def __init__(self, X, y, topk=10, missing=None, metric='auc', objective='binary:logistic', fix_params={},
                 n_jobs=8, opt_seed=None):
        """
        :param objective: 'binary:logistic', 'multi:softmax', 'reg:linear'
        """
        self.data = xgb.DMatrix(X, y, missing=missing)
        self.topk = topk
        self.metric = metric
        self.objective = objective

        self.fix_params = fix_params  # 固定不需要调节的参数
        self.opt_seed = opt_seed
        self.n_jobs = n_jobs

        self.params_ls = []
        self.params_ls_sk = []
        self.params_best = {}
        self.params_best_sk = {}
        self.params_opt_df = None

        self._iter_ls = []

        if self.fix_params:
            print('Fixed Params: \033[94m%s\033[0m\n' % self.fix_params)

    @property
    def best_model(self):
        if self.params_best:
            return xgb.train(dtrain=self.data, **self.params_best)
        else:
            print('\033[94m%s\033[0m\n' % "Please Run !")

    def run(self, n_iter=5, save_log=False):

        BoParams = {
            'max_depth': (5, 16),
            'min_child_weight': (1, 10),
            'gamma': (0, 1),
            'subsample': (0.6, 1),
            'colsample_bytree': (0.6, 1),
            'reg_alpha': (0, 1),
            'reg_lambda': (0, 1),
        }
        optimizer = BayesianOptimization(self.__evaluator, BoParams, self.opt_seed)
        if save_log:
            logger = JSONLogger(path="./opt_xgb_logs.json")
            optimizer.subscribe(Events.OPTMIZATION_STEP, logger)

        if self.fix_params:
            optimizer.set_bounds({k: (v, v) for k, v in self.fix_params.items()})

        # optimizer.probe(
        #     {'max_depth': 7,
        #      'min_child_weight': 1,
        #      'gamma': 0,
        #      'subsample': 0.8,
        #      'colsample_bytree': 0.8,
        #      'reg_alpha': 0.01,
        #      'reg_lambda': 1})

        gp_params = {"alpha": 1e-5, "n_restarts_optimizer": 3}
        optimizer.maximize(init_points=3, n_iter=n_iter, acq='ucb', kappa=2.576, xi=0.0, **gp_params)
        self.__get_params(optimizer)

    def __evaluator(self, max_depth, gamma, min_child_weight, subsample, colsample_bytree,
                    reg_alpha, reg_lambda):
        self.__params_sk = dict(
            silent=True,
            booster='gbtree',
            objective=self.objective,
            max_depth=int(max_depth),
            learning_rate=0.01,
            gamma=gamma,  # min_split_gain
            min_child_weight=min_child_weight,
            subsample=subsample,
            colsample_bylevel=0.8,  # 每一层的列数
            colsample_bytree=colsample_bytree,
            reg_alpha=reg_alpha,
            reg_lambda=reg_lambda,
            scale_pos_weight=1,
            random_state=0,
            n_jobs=self.n_jobs
        )
        params = self.__params_sk.copy()
        params['eta'] = params.pop('learning_rate')
        params['alpha'] = params.pop('reg_alpha')
        params['lambda'] = params.pop('reg_lambda')
        params['eval_metric'] = self.metric

        _ = xgb.cv(params,
                   self.data,
                   num_boost_round=3600,
                   nfold=5,
                   early_stopping_rounds=100,
                   show_stdv=False,
                   stratified=False if 'reg' in self.objective else True,
                   as_pandas=False)['test-%s-mean' % self.metric]
        self._iter_ls.append(len(_))
        return -_[-1] if 'reg' in self.objective else _[-1]

    def __get_params(self, optimizer):
        self.params_opt_df = (
            pd.concat([pd.DataFrame({'iter': self._iter_ls}), pd.DataFrame(optimizer.res)], 1)
                .sort_values('target', ascending=False)
                .reset_index(drop=True)[:self.topk])

        for _, (i, p, _) in self.params_opt_df.iterrows():
            params_sk = {**self.__params_sk, **p, **{'n_estimators': i}}

            params_sk['max_depth'] = int(params_sk['max_depth'])
            params_sk = {k: float('%.3f' % v) if isinstance(v, float) else v for k, v in params_sk.items()}
            self.params_ls_sk.append(params_sk)

            params = params_sk.copy()
            params['eta'] = params.pop('learning_rate')
            params['alpha'] = params.pop('reg_alpha')
            params['lambda'] = params.pop('reg_lambda')
            num_boost_round = params.pop('n_estimators')
            self.params_ls.append({'params': params, 'num_boost_round': num_boost_round})

        self.params_best = self.params_ls[0]
        self.params_best_sk = self.params_ls_sk[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'emoji_feats'
__author__ = 'JieYuan'
__mtime__ = '2019/4/16'
"""

import emoji
from collections import Counter


class Emoji(object):

    def __init__(self):
        self.regexp = emoji.get_emoji_regexp()

    def counter(self, s="👍💝💝😁😁😁 言之有理", topk=None):
        c = Counter(self.regexp.findall(s))
        return c.most_common(topk)


if __name__ == '__main__':
    e = Emoji()
    print(e.counter())
    print(e.counter(topk=2))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'reduce_mem_usage'
__author__ = 'JieYuan'
__mtime__ = '19-1-31'
"""
import numpy as np
from tqdm import tqdm


# reduce memory
def reduce_mem_usage(df):
    numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
    start_mem = df.memory_usage().sum() / 1024 ** 2
    for col in tqdm(df.columns):
        col_type = df[col].dtypes
        if col_type in numerics:
            c_min = df[col].min()
            c_max = df[col].max()
            if str(col_type)[:3] == 'int':
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)
            else:
                if c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                    df[col] = df[col].astype(np.float16)
                elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)

    end_mem = df.memory_usage().sum() / 1024 ** 2
    print('Memory usage after optimization is: {:.2f} MB'.format(end_mem))
    print('Decreased by {:.1f}%'.format(100 * (start_mem - end_mem) / start_mem))

    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = '__init__.py'
__author__ = 'JieYuan'
__mtime__ = '18-12-17'
"""

from .string import cprint, cstring

group_by_step = lambda ls, step=3: [ls[idx: idx + step] for idx in range(0, len(ls), step)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : pandas
# @Time         : 2019-06-24 13:20
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


# https://github.com/tensorflow/tensorflow/issues/20995
import tensorflow as tf

# input_func = tf.estimator.inputs.pandas_input_fn()
#
#
# tf.estim


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
__title__ = 'rgf'
__author__ = 'JieYuan'
__mtime__ = '19-1-2'
"""

from sklearn.model_selection import GridSearchCV
from rgf import RGFClassifier

rgf = RGFClassifier(
    max_leaf=1000,
    algorithm="RGF",
    loss="Log",
    l2=0.01,
    sl2=0.01,
    normalize=False,
    min_samples_leaf=10,
    n_iter=None,
    opt_interval=100,
    learning_rate=.5,
    calc_prob="sigmoid",
    n_jobs=8,
    memory_policy="generous"
)

parameters = {
    'max_leaf': [1000, 1200],
    'l2': [0.01, 0.1, 0.2, 0.3],
    'min_samples_leaf': [5, 10]
}

clf = GridSearchCV(
    estimator=rgf,
    param_grid=parameters,
    scoring='roc_auc',
    n_jobs=3,
    verbose=1,
    cv=3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : iWork.
# @File         : DateTimeFeat
# @Time         : 2019-06-13 23:16
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import pandas as pd


class DateTimeFeats(object):
    """
    pandas 时间/日期特征工程
    常见格式：
        1. 时间戳
        2. 时间字符串

    # 很强大: 解析不出为空
    pd.to_datetime(ts, 'coerce', unit='s', infer_datetime_format=True)
    """

    def __init__(self):
        self.time_granularity = ("year", "quarter", "month",
                                 "day", "hour", "minute",
                                 "week", "weekday", "weekofyear")

    def transform(self, s: pd.Series, include=None):
        feats = include if include else self.time_granularity

        if isinstance(s[0], (int, float)):  # 时间戳
            ts = self.timestamp2date(s)
        else:
            ts = self.dateStr2date(s)

        ts = ts.map(lambda t: list(self._func(t, feats)))
        return pd.DataFrame(ts.tolist(), columns=feats)

    def _func(self, t, feats):
        for feat in feats:
            _ = t.__getattribute__(feat)
            if callable(_):
                yield _()
            else:
                yield _

    def timestamp2date(self, ts):
        return pd.to_datetime(ts, 'coerce', unit='s', infer_datetime_format=True)

    def dateStr2date(self, ts):
        try:
            _ = ts.astype('datetime64[ns]')
        except Exception as e:
            print("astype('datetime64[ns]'): %s" % e)
            _ = pd.to_datetime(ts, 'coerce', infer_datetime_format=True)
        return _


if __name__ == '__main__':
    import pandas as pd

    ts = pd.Series([pd.datetime.today()] * 10)
    print(DateTimeFeats().transform(ts))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : losses
# @Time         : 2019-07-03 17:15
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import tensorflow as tf
from tensorflow.python.keras import backend as K


# 参考losses.py编写自己的损失函数
# Retinanet 针对正负样本不均衡问题设计的损失函数focal_loss
# Keras 以tf为后端；可以使用Keras的math方法或者tf的math方法写都可以keras被识别
def focal_loss(y_true, y_pred, gamma=2., alpha=.25):
    pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
    pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
    return K.mean(-alpha * K.pow(1 - pt_1, gamma) * K.log(pt_1) - (1 - alpha) * K.pow(pt_0, gamma) * K.log(1 - pt_0),
                  axis=-1)

    # return -K.mean(alpha * K.pow(1. - pt_1, gamma) * K.log(pt_1)) - K.mean(
    #     (1 - alpha) * K.pow(pt_0, gamma) * K.log(1. - pt_0))


# return -K.sum(alpha * K.pow(1. - pt_1, gamma) * K.log(K.epsilon() + pt_1)) - K.sum(
#     (1 - alpha) * K.pow(pt_0, gamma) * K.log(1. - pt_0 + K.epsilon()))

# https://github.com/umbertogriffo/focal-loss-keras/blob/master/losses.py
# Define our custom loss function
def focal_loss(y_true, y_pred):
    gamma = 2.0
    alpha = 0.25
    pt_1 = tf.where(tf.equal(y_true, 1), y_pred, tf.ones_like(y_pred))
    pt_0 = tf.where(tf.equal(y_true, 0), y_pred, tf.zeros_like(y_pred))
    return - K.sum(alpha * K.pow(1. - pt_1, gamma) * K.log(pt_1)) - K.sum(
        (1 - alpha) * K.pow(pt_0, gamma) * K.log(1. - pt_0))


# 未调试
def catergorical_focal_loss(gamma=2.0, alpha=0.25):
    """
    Implementation of Focal Loss from the paper in multiclass classification
    Formula:
        loss = -alpha*((1-p_t)^gamma)*log(p_t)
    Parameters:
        alpha -- the same as wighting factor in balanced cross entropy
        gamma -- focusing parameter for modulating factor (1-p)
    Default value:
        gamma -- 2.0 as mentioned in the paper
        alpha -- 0.25 as mentioned in the paper
    """

    def focal_loss(y_true, y_pred):
        # Define epsilon so that the backpropagation will no result in NaN
        # for o divisor case
        epsilon = K.epsilon()
        # Add the epsilon to prediction value
        # y_pred = y_pred + epsilon
        # Clip the prediction value
        y_pred = K.clip(y_pred, epsilon, 1.0 - epsilon)
        # Calculate cross entropy
        cross_entropy = -y_true * K.log(y_pred)

        # Calculate weight that consists of modulating factor and weighting factor
        weight = alpha * y_true * K.pow((1 - y_pred), gamma)
        # Calculate focal loss
        loss = weight * cross_entropy
        # Sum the losses in mini_batch
        loss = K.sum(loss, axis=1)
        return loss

    return focal_loss


def binary_focal_loss(gamma=2.0, alpha=0.25):
    """
        Implementation of Focal Loss from the paper in multiclass classification
        Formula:
            loss = -alpha_t*((1-p_t)^gamma)*log(p_t)

            p_t = y_pred, if y_true = 1
            p_t = 1-y_pred, otherwise

            alpha_t = alpha, if y_true=1
            alpha_t = 1-alpha, otherwise

            cross_entropy = -log(p_t)
        Parameters:
            alpha -- the same as wighting factor in balanced cross entropy
            gamma -- focusing parameter for modulating factor (1-p)
        Default value:
            gamma -- 2.0 as mentioned in the paper
            alpha -- 0.25 as mentioned in the paper
        """

    def focal_loss(y_true, y_pred):
        # Define espislon so that the backpropagation will not result int NaN
        # for 0 divisor case
        epsilon = K.epsilon()
        # Add the epsilon to prediction value
        # y_pred = y_pred + epsilon
        # Clip the prediction value
        y_pred = K.clip(y_pred, epsilon, 1.0 - epsilon)

        alpha_factor = K.ones_like(y_true) * alpha

        # Calculate p_t
        p_t = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)

        # Calculate alpha_t
        alpha_t = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)
        # Calculate cross entropy
        cross_entropy = -K.log(p_t)
        weight = alpha_t * K.pow((1 - p_t), gamma)
        # Calculate focal loss
        loss = weight * cross_entropy
        # Sum the losses in mini_batch
        loss = K.sum(loss, axis=1)

        return loss

    return focal_loss



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : target
# @Time         : 2019-09-14 18:36
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


def target_distributed(df, by, target_name='label'):
    _ = df.groupby(by)[target_name].value_counts(1)
    return _



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : TextCNN
# @Time         : 2019-06-21 18:56
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from .BaseModel import BaseModel
from tensorflow.python.keras import Input, Model
from tensorflow.python.keras.layers import Dense, Concatenate, Dropout
from tensorflow.python.keras.layers import Conv1D, GlobalMaxPool1D, GlobalAvgPool1D


class TextCNN(BaseModel):
    """ TextCNN:
    1. embedding layers: embeddings = model.layers[0].get_weights()[0]
    2. convolution layer,
    3. max-pooling,
    4. softmax layer.
    数据量较大：可以直接随机初始化embeddings，然后基于语料通过训练模型网络来对embeddings进行更新和学习。
    数据量较小：可以利用外部语料来预训练(pre-train)词向量，然后输入到Embedding层，用预训练的词向量矩阵初始化embeddings。（通过设置weights=[embedding_matrix]）。
    静态(static)方式：训练过程中不再更新embeddings。实质上属于迁移学习，特别是在目标领域数据量比较小的情况下，采用静态的词向量效果也不错。（通过设置trainable=False）
    非静态(non-static)方式：在训练过程中对embeddings进行更新和微调(fine tune)，能加速收敛。（通过设置trainable=True）
    """

    def __init__(self, max_tokens, maxlen=128, embedding_size=None, num_class=1, kernel_size_list=(3, 4, 5),
                 weights=None):
        """

        :param embedding_size: 类别/实体嵌入时可不指定

        model = TextCNN(max_token, maxlen, num_class=1)()
        model.compile('adam', 'binary_crossentropy', metrics=['accuracy'])
        model.fit_generator(DataIter(X, y), epochs=5)
        """
        super().__init__(max_tokens, maxlen, embedding_size, num_class, weights)
        self.kernel_size_list = kernel_size_list

    def get_model(self):
        input = Input(self.maxlen)
        # Embedding part can try multichannel as same as origin paper
        embedding = self.embedding_layer(input)
        convs = []
        for kernel_size in self.kernel_size_list:
            c = Conv1D(128, kernel_size, activation='relu')(embedding)  # 卷积
            # c = Dropout(0.5)(c)
            p = GlobalMaxPool1D()(c)  # 池化
            # p = GlobalAvgPool1D()(c)
            convs.append(p)
        x = Concatenate()(convs)
        output = Dense(self.num_class, activation=self.last_activation)(x)

        model = Model(inputs=input, outputs=output)
        return model


if __name__ == '__main__':
    TextCNN(1000, 10)()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : tql-Python.
# @File         : plot_set
# @Time         : 2019-06-20 11:28
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import matplotlib.pyplot as plt


def set_plot():
    """
    plt.rcParams['font.sans-serif'] = ['Simhei']  # 中文乱码的处理
    plt.rcParams['font.family'] = 'sans-serif'
    plt.rcParams['axes.unicode_minus'] = False  # 负号
    plt.rcParams["text.usetex"] = False
    plt.rcParams["legend.numpoints"] = 1
    plt.rcParams["figure.figsize"] = (18, 9)  # (12, 6)
    plt.rcParams["figure.dpi"] = 128
    plt.rcParams["savefig.dpi"] = plt.rcParams["figure.dpi"]
    plt.rcParams["font.size"] = 12
    plt.rcParams["pdf.fonttype"] = 42
    """
    plt.rcParams['font.sans-serif'] = ['Simhei']  # 中文乱码的处理
    plt.rcParams['font.family'] = 'sans-serif'
    plt.rcParams['axes.unicode_minus'] = False  # 负号
    plt.rcParams["text.usetex"] = False
    plt.rcParams["legend.numpoints"] = 1
    plt.rcParams["figure.figsize"] = (18, 9)  # (12, 6)
    plt.rcParams["figure.dpi"] = 128
    plt.rcParams["savefig.dpi"] = plt.rcParams["figure.dpi"]
    plt.rcParams["font.size"] = 12
    plt.rcParams["pdf.fonttype"] = 42
    print('Setting Success!')



#!/usr/bin/env python
#-*- coding: utf-8 -*-
"""
__title__ = 'lag'
__author__ = 'JieYuan'
__mtime__ = '19-2-14'
"""

def lag(df, by_name, col_name, n=1):
    df = df.copy()
    df[col_name + '_' + str(n)] = df.groupby(by_name)[col_name].shift(n)
    return df